@@ -21,9 +21,12 @@ import java.io.File;
 import java.io.IOException;
 import java.io.InputStream;
 import java.io.RandomAccessFile;
+import java.util.Arrays;
 import java.util.Collections;
+import java.util.Comparator;
 import java.util.Enumeration;
 import java.util.HashMap;
+import java.util.LinkedHashMap;
 import java.util.Map;
 import java.util.zip.Inflater;
 import java.util.zip.InflaterInputStream;
@@ -72,7 +75,7 @@ public class ZipFile {
      * Maps ZipArchiveEntrys to Longs, recording the offsets of the local
      * file headers.
      */
-    private final Map entries = new HashMap(HASH_SIZE);
+    private final Map entries = new LinkedHashMap(HASH_SIZE);
 
     /**
      * Maps String to ZipArchiveEntrys, name -> actual entry.
@@ -98,6 +101,11 @@ public class ZipFile {
      */
     private final ZipEncoding zipEncoding;
 
+    /**
+     * File name of actual source.
+     */
+    private final String archiveName;
+
     /**
      * The actual data source.
      */
@@ -108,6 +116,11 @@ public class ZipFile {
      */
     private final boolean useUnicodeExtraFields;
 
+    /**
+     * Whether the file is closed.
+     */
+    private boolean closed;
+
     /**
      * Opens the given file for reading, assuming "UTF8" for file names.
      *
@@ -172,20 +185,22 @@ public class ZipFile {
      */
     public ZipFile(File f, String encoding, boolean useUnicodeExtraFields)
         throws IOException {
+        this.archiveName = f.getAbsolutePath();
         this.encoding = encoding;
         this.zipEncoding = ZipEncodingHelper.getZipEncoding(encoding);
         this.useUnicodeExtraFields = useUnicodeExtraFields;
         archive = new RandomAccessFile(f, "r");
         boolean success = false;
         try {
-            Map entriesWithoutEFS = populateFromCentralDirectory();
-            resolveLocalFileHeaderData(entriesWithoutEFS);
+            Map entriesWithoutUTF8Flag = populateFromCentralDirectory();
+            resolveLocalFileHeaderData(entriesWithoutUTF8Flag);
             success = true;
         } finally {
             if (!success) {
                 try {
+                    closed = true;
                     archive.close();
-                } catch (IOException e2) {
+                } catch (IOException e2) { // NOPMD
                     // swallow, throw the original exception instead
                 }
             }
@@ -206,6 +221,11 @@ public class ZipFile {
      * @throws IOException if an error occurs closing the archive.
      */
     public void close() throws IOException {
+        // this flag is only written here and read in finalize() which
+        // can never be run in parallel.
+        // no synchronization needed.
+        closed = true;
+
         archive.close();
     }
 
@@ -218,20 +238,40 @@ public class ZipFile {
         if (zipfile != null) {
             try {
                 zipfile.close();
-            } catch (IOException e) {
-                //ignore
+            } catch (IOException e) { // NOPMD
+                //ignore, that's why the method is called "quietly"
             }
         }
     }
 
     /**
      * Returns all entries.
+     *
+     * <p>Entries will be returned in the same order they appear
+     * within the archive's central directory.</p>
+     *
      * @return all entries as {@link ZipArchiveEntry} instances
      */
     public Enumeration getEntries() {
         return Collections.enumeration(entries.keySet());
     }
 
+    /**
+     * Returns all entries in physical order.
+     *
+     * <p>Entries will be returned in the same order their contents
+     * appear within the archive.</p>
+     *
+     * @return all entries as {@link ZipArchiveEntry} instances
+     *
+     * @since Commons Compress 1.1
+     */
+    public Enumeration getEntriesInPhysicalOrder() {
+        Object[] allEntries = entries.keySet().toArray();
+        Arrays.sort(allEntries, OFFSET_COMPARATOR);
+        return Collections.enumeration(Arrays.asList(allEntries));
+    }
+
     /**
      * Returns a named entry - or <code>null</code> if no entry by
      * that name exists.
@@ -243,13 +283,24 @@ public class ZipFile {
         return (ZipArchiveEntry) nameMap.get(name);
     }
 
+    /**
+     * Whether this class is able to read the given entry.
+     *
+     * <p>May return false if it is set up to use encryption or a
+     * compression method that hasn't been implemented yet.</p>
+     * @since Apache Commons Compress 1.1
+     */
+    public boolean canReadEntryData(ZipArchiveEntry ze) {
+        return ZipUtil.canHandleEntryData(ze);
+    }
+
     /**
      * Returns an InputStream for reading the contents of the given entry.
+     *
      * @param ze the entry to get the stream for.
      * @return a stream to read the entry from.
      * @throws IOException if unable to create an input stream from the zipenty
-     * @throws ZipException if the zipentry has an unsupported
-     * compression method
+     * @throws ZipException if the zipentry uses an unsupported feature
      */
     public InputStream getInputStream(ZipArchiveEntry ze)
         throws IOException, ZipException {
@@ -257,6 +308,7 @@ public class ZipFile {
         if (offsetEntry == null) {
             return null;
         }
+        ZipUtil.checkRequestedFeatures(ze);
         long start = offsetEntry.dataOffset;
         BoundedInputStream bis =
             new BoundedInputStream(start, ze.getCompressedSize());
@@ -265,13 +317,36 @@ public class ZipFile {
                 return bis;
             case ZipArchiveEntry.DEFLATED:
                 bis.addDummy();
-                return new InflaterInputStream(bis, new Inflater(true));
+                final Inflater inflater = new Inflater(true);
+                return new InflaterInputStream(bis, inflater) {
+                    public void close() throws IOException {
+                        super.close();
+                        inflater.end();
+                    }
+                };
             default:
                 throw new ZipException("Found unsupported compression method "
                                        + ze.getMethod());
         }
     }
 
+    /**
+     * Ensures that the close method of this zipfile is called when
+     * there are no more references to it.
+     * @see #close()
+     */
+    protected void finalize() throws Throwable {
+        try {
+            if (!closed) {
+                System.err.println("Cleaning up unclosed ZipFile for archive "
+                                   + archiveName);
+                close();
+            }
+        } finally {
+            super.finalize();
+        }
+    }
+
     private static final int CFH_LEN =
         /* version made by                 */ SHORT
         /* version needed to extract       */ + SHORT
@@ -304,7 +379,7 @@ public class ZipFile {
      */
     private Map populateFromCentralDirectory()
         throws IOException {
-        HashMap noEFS = new HashMap();
+        HashMap noUTF8Flag = new HashMap();
 
         positionAtCentralDirectory();
 
@@ -329,11 +404,11 @@ public class ZipFile {
 
             off += SHORT; // skip version info
 
-            final int generalPurposeFlag = ZipShort.getValue(cfh, off);
-            final boolean hasEFS = 
-                (generalPurposeFlag & ZipArchiveOutputStream.EFS_FLAG) != 0;
+            final GeneralPurposeBit gpFlag = GeneralPurposeBit.parse(cfh, off);
+            final boolean hasUTF8Flag = gpFlag.usesUTF8ForNames();
             final ZipEncoding entryEncoding =
-                hasEFS ? ZipEncodingHelper.UTF8_ZIP_ENCODING : zipEncoding;
+                hasUTF8Flag ? ZipEncodingHelper.UTF8_ZIP_ENCODING : zipEncoding;
+            ze.setGeneralPurposeBit(gpFlag);
 
             off += SHORT;
 
@@ -375,7 +450,7 @@ public class ZipFile {
 
             byte[] fileName = new byte[fileNameLen];
             archive.readFully(fileName);
-            ze.setName(entryEncoding.decode(fileName));
+            ze.setName(entryEncoding.decode(fileName), fileName);
 
             // LFH offset,
             OffsetEntry offset = new OffsetEntry();
@@ -396,11 +471,11 @@ public class ZipFile {
             archive.readFully(signatureBytes);
             sig = ZipLong.getValue(signatureBytes);
 
-            if (!hasEFS && useUnicodeExtraFields) {
-                noEFS.put(ze, new NameAndComment(fileName, comment));
+            if (!hasUTF8Flag && useUnicodeExtraFields) {
+                noUTF8Flag.put(ze, new NameAndComment(fileName, comment));
             }
         }
-        return noEFS;
+        return noUTF8Flag;
     }
 
     private static final int MIN_EOCD_SIZE =
@@ -495,7 +570,7 @@ public class ZipFile {
      * <p>Also records the offsets for the data to read from the
      * entries.</p>
      */
-    private void resolveLocalFileHeaderData(Map entriesWithoutEFS)
+    private void resolveLocalFileHeaderData(Map entriesWithoutUTF8Flag)
         throws IOException {
         Enumeration e = getEntries();
         while (e.hasMoreElements()) {
@@ -527,9 +602,9 @@ public class ZipFile {
             offsetEntry.dataOffset = offset + LFH_OFFSET_FOR_FILENAME_LENGTH
                 + SHORT + SHORT + fileNameLen + extraFieldLen;
 
-            if (entriesWithoutEFS.containsKey(ze)) {
+            if (entriesWithoutUTF8Flag.containsKey(ze)) {
                 String orig = ze.getName();
-                NameAndComment nc = (NameAndComment) entriesWithoutEFS.get(ze);
+                NameAndComment nc = (NameAndComment) entriesWithoutUTF8Flag.get(ze);
                 ZipUtil.setNameAndCommentFromExtraFields(ze, nc.name,
                                                          nc.comment);
                 if (!orig.equals(ze.getName())) {
@@ -631,4 +706,34 @@ public class ZipFile {
             this.comment = comment;
         }
     }
+
+    /**
+     * Compares two ZipArchiveEntries based on their offset within the archive.
+     *
+     * <p>Won't return any meaningful results if one of the entries
+     * isn't part of the archive at all.</p>
+     *
+     * @since Commons Compress 1.1
+     */
+    private final Comparator OFFSET_COMPARATOR =
+        new Comparator() {
+            public int compare(Object o1, Object o2) {
+                if (o1 == o2)
+                    return 0;
+
+                ZipArchiveEntry e1 = (ZipArchiveEntry) o1;
+                ZipArchiveEntry e2 = (ZipArchiveEntry) o2;
+
+                OffsetEntry off1 = (OffsetEntry) entries.get(e1);
+                OffsetEntry off2 = (OffsetEntry) entries.get(e2);
+                if (off1 == null) {
+                    return 1;
+                }
+                if (off2 == null) {
+                    return -1;
+                }
+                long val = (off1.headerOffset - off2.headerOffset);
+                return val == 0 ? 0 : val < 0 ? -1 : +1;
+            }
+        };
 }
